# commands/hccnstable.py

import ast
import pandas as pd
from sklearn.metrics import adjusted_rand_score as adjustedRandIndex
from sklearn.metrics import adjusted_mutual_info_score as AMI
import os
import numpy as np
from Bio import Phylo

def add_hccnstable_subparser(subparsers):
    """
    Add the hccnstable subcommand to the CLI tool.

    This subcommand is used to calculate Evolutionary CN Stability.
    """
    parser = subparsers.add_parser(
        "hccnstable",
        help="Calculate the ACC of Evolutionary CN Stability."
    )
    parser.add_argument(
        "--tool-maternal-cna-files", 
        type=str, 
        nargs = "+",
        required=True, 
        metavar="",
        help="Paths to the maternal CNA files generated by different tools, separated by spaces"
    )
    parser.add_argument(
        "--tool-paternal-cna-files", 
        type=str, 
        nargs = "+",
        required=True, 
        metavar="",
        help="Paths to the paternal CNA files generated by different tools, separated by spaces"
    )
    parser.add_argument(
        "--tool-names", 
        type=str, 
        nargs = "+",
        required=True, 
        metavar="",
        help="List of tool names corresponding to the maternal and paternal CNA files provided."
    )
    parser.add_argument(
        "--changes-file", 
        type=str, 
        required=True, 
        metavar="",
        help="Path to the changes file that contains information about changes in CN status."
    )
    parser.add_argument(
        "--tree-file", 
        type=str, 
        required=True, 
        metavar="",
        help="Path to the tree file that provides the hierarchical clonal structure."
    )
    parser.add_argument(
        "--output-dir", 
        type=str, 
        default="./output", 
        metavar="",
        help="Path to the file where the results will be saved, default: './output' "
    )
    return parser


def is_descendant_or_self(tree, ancestor_name, descendant_name):
    """
    Check if a node is a descendant of or identical to an ancestor node in the tree.

    Parameters:
    tree: Tree object to search.
    ancestor_name (str): Name of the ancestor node.
    descendant_name (str): Name of the descendant node.

    Returns:
    bool: True if the node is a descendant of or identical to the ancestor, otherwise False.
    """
    ancestor = tree.find_any(name=ancestor_name)
    descendant = tree.find_any(name=descendant_name)
    
    if ancestor is None or descendant is None:
        return False  
    if ancestor == descendant :
        return True

    for clade in ancestor.find_clades():
        if clade == descendant:
            return True
            
    return False

def get_descendant(tree,ancestor_name):
    """
    Retrieve the names of all descendant nodes of a given ancestor in the tree.

    Parameters:
    tree: Tree object to search.
    ancestor_name (str): Name of the ancestor node.

    Returns:
    list: List of names of descendant nodes.
    """
    ancestor = tree.find_any(name=ancestor_name)
    descendant = ancestor.find_clades()
            
    descendant_names = [clade.name for clade in descendant if clade.name is not None]
    
    return descendant_names

def add_check_list(tree,change_data):
    """
    Add a 'check_child_list' column to the change data for nodes requiring verification.

    Parameters:
    tree: Tree object containing hierarchical relationships.
    change_data (DataFrame): DataFrame with change data.

    Returns:
    DataFrame: Updated DataFrame with 'check_child_list' column.
    """

    if 'check_child_list' not in change_data.columns:
        change_data['check_child_list'] = None  
    change_data = change_data.astype({'check_child_list': 'object'})  
    for index,row in change_data.iterrows():
        start = row['Child'] 
        segment = row['Segment'] 
        haplotype = row['Haplotype']
        descendant = get_descendant(tree,start)
        sub_clone = change_data[(change_data['Segment']==segment) & (change_data['Haplotype'] == haplotype)]['Child'].unique()

        del_clone_list = []
        for clone in sub_clone:
            if (clone == start):
                continue
            elif is_descendant_or_self(tree, start,clone) :
                sub_descendant = list(get_descendant(tree, clone))
                del_clone_list += sub_descendant


        update_descendant = list(set(descendant) - set(del_clone_list))
        change_data.at[index, 'check_child_list'] = update_descendant

    return change_data


def calculate_mode_for_columns(row, columns_prefix, df):
    """
    Calculate the mode for a subset of columns matching a specific prefix.

    Parameters:
    row (Series): A row from the DataFrame.
    columns_prefix (str): Prefix to filter relevant columns.
    df (DataFrame): DataFrame containing the data.

    Returns:
    Any: The most frequent value (mode) or None if no values exist.
    """
    columns = [col for col in df.columns if col.startswith(f"{columns_prefix}_")]
    if columns:
        values = row[columns]
        mode_value = values.mode()
        return mode_value.iloc[0] if not mode_value.empty else None
    else:
        return None

def process_haplotype_result(row, predict_data, haplotype, change_data, index):
    """
    Process a single row of change data for a specific haplotype.

    Parameters:
    row (Series): Row of the change_data DataFrame.
    prediction_data (DataFrame): DataFrame containing prediction results.
    haplotype (str): Haplotype being processed ('maternal' or 'paternal').
    change_data (DataFrame): DataFrame containing change data.
    index (int): Index of the current row in change_data.
    """
    change_result = int(row['Change'].split('->')[1])
    region = row['Segment']

    tool_row = predict_data[predict_data['region'] == region]

    if tool_row.empty:
        return

    tool_row = tool_row.iloc[0]
    change_data.loc[index, 'result'] = 1
    clone_list = row['check_child_list']
    if isinstance(clone_list, str):
        clone_list = ast.literal_eval(clone_list)

    for clone in clone_list:
        clone_result = calculate_mode_for_columns(tool_row,clone,predict_data)
        if clone_result != change_result:
            change_data.loc[index, 'result'] = 0
    

def process_change_data(change_data, maternal_result, paternal_result):
    for index, row in change_data.iterrows():
        if row['Haplotype'] == 'maternal':
            process_haplotype_result(row, maternal_result, 'maternal', change_data, index)
        elif row['Haplotype'] == 'paternal':
            process_haplotype_result(row, paternal_result, 'paternal', change_data, index)


# python src/main.py hccnstable 
# --tool-maternal-cna-files data/chisel_5x_maternal_cnv.csv 
# --tool-paternal-cna-files data/chisel_5x_paternal_cnv.csv 
# --tool-names chisel_5x --changes-file profile/changes.csv 
# --tree-file profile/tree.newick

def run(args):
    # Validate input
    if (len(args.tool_maternal_cna_files) != len(args.tool_names) ) and (len(args.tool_maternal_cna_files) != len(args.tool_paternal_cna_files)):
        raise ValueError("The number of classification files must match the number of tool names.")
    
    
    os.makedirs(args.output_dir, exist_ok=True)

    tree = Phylo.read(args.tree_file, "newick")
    change_data = pd.read_csv(args.changes_file)
    change_data = add_check_list(tree,change_data)

    results = []
    for tool_maternal_path,tool_paternal_path, tool_name in zip(args.tool_maternal_cna_files, args.tool_paternal_cna_files,args.tool_names):

        predict_maternal = pd.read_csv(tool_maternal_path)
        predict_paternal = pd.read_csv(tool_paternal_path)

        process_change_data(change_data, predict_maternal, predict_paternal)

        for type in change_data['Type'].unique().tolist():

            acc = change_data[change_data['Type']==type]['result'].mean()

            results.append({
                "Tool": tool_name,
                "Type": type,
                "ACC": acc
            })
    
    result_df = pd.DataFrame(results)
    output_file = os.path.join(args.output_dir, "evolution_cn_stability_acc.csv")
    result_df.to_csv(output_file, index=False)
    print(f"Results saved to {output_file}")



# commands/subdetect.py

import pandas as pd
import os
from sklearn.metrics import adjusted_rand_score as adjustedRandIndex, mean_squared_error
from sklearn.metrics import adjusted_mutual_info_score as AMI


def add_subdetect_subparser(subparsers):
    """
    Add the subdetect subcommand to the CLI tool.

    This subcommand is used to calculate the Adjusted Mutual Information (AMI)
    and Adjusted Rand Index (ARI) based on a classification file.
    """
    parser = subparsers.add_parser(
        "subdetect",
        help="Calculate AMI (Adjusted Mutual Information) and ARI (Adjusted Rand Index)."
    )
    parser.add_argument(
        "--classification-files", 
        type=str, 
        nargs = "+",
        required=True, 
        metavar="",
        help="Paths to the classification files, separated by spaces. Each file must "
            "contain 'cell_id' and 'clone_id' columns. The 'clone_id' column represents "
            "classification information, while 'cell_id' combines 'ground_truth_clone_id' "
            "and cell details using an underscore."
    )
    parser.add_argument(
        "--tool-names", 
        type=str, 
        nargs = "+",
        required=True, 
        metavar="",
        help="List of tool names corresponding to the classification files"
    )
    parser.add_argument(
        "--output-dir", 
        type=str, 
        default="./output", 
        metavar="",
        help="Path to the file where the results will be saved, default: './output' "
    )
    return parser

def evaluate_clustering_results(file_path, tool_name="tool_name"):
    """
    Process a CSV file containing clustering data and evaluate clustering performance using ARI and AMI metrics.

    Parameters:
    file_path (str): Path to the input CSV file containing clustering data.
    tool_name (str, optional): Name of the tool used for generating the data. Defaults to "tool_name".

    Returns:
    dict: A dictionary containing the tool name, Adjusted Rand Index (ARI), and Adjusted Mutual Information (AMI).
    """
    data = pd.read_csv(file_path)
    
    # Extract the 'clone_id' column as clusters
    clusters = data['clone_id']
    
    # Extract 'cell_id', split by "_" to get cell clones, and to generate numeric labels
    cell_clones = pd.factorize(data['cell_id'].str.split("_").str[0])[0]
    
    # Compute the Adjusted Rand Index (ARI) between clusters and cell clones
    ari = adjustedRandIndex(clusters, cell_clones)
    
    # Compute the Adjusted Mutual Information (AMI) between clusters and cell clones
    ami = AMI(clusters, cell_clones)
    
    # Return the results as a dictionary
    return {'Tool': tool_name, 'ARI': ari, 'AMI': ami}

# python src/main.py subdetect --classification-files  data\clustering_result.csv  --tool-names chisel

def run(args):
    # Validate input
    if len(args.classification_files) != len(args.tool_names):
        raise ValueError("The number of classification files must match the number of tool names.")
    
    os.makedirs(args.output_dir, exist_ok=True)

    results = []
    for path, tool in zip(args.classification_files, args.tool_names):
        result = evaluate_clustering_results(path, tool)
        results.append(result)
    
    result_df = pd.DataFrame(results)
    output_file = os.path.join(args.output_dir, "AMI_ARI_results.csv")
    result_df.to_csv(output_file, index=False)
    print(f"Results saved to {output_file}")



# commands/cndetect.py

import pandas as pd
import numpy as np
import os

from sklearn.isotonic import spearmanr
from sklearn.metrics import mean_squared_error


def add_cndetect_subparser(subparsers):
    """
    Add the cndetect subcommand to the CLI tool.

    This subcommand is used to calculate the RMSE,ACC,SCC of copy number (CN) status at the bin level.
    """
    parser = subparsers.add_parser(
        "cndetect",
        help="Calculate the RMSE,ACC,SCC of CN status at the bin level."
    )
    parser.add_argument(
        "--tool-cna-files", 
        type=str, 
        nargs = "+",
        required=True, 
        metavar="",
        help="Paths to the CNA files generated by different tools. "
    )
    parser.add_argument(
        "--cna-profile-file", 
        type=str, 
        required=True, 
        metavar="",
        help="Path to the ground truth CNA profile file."
    )
    parser.add_argument(
        "--tool-names", 
        type=str, 
        nargs = "+",
        required=True, 
        metavar="",
        help="List of tool names corresponding to the classification files"
    )
    parser.add_argument(
        "--output-dir", 
        type=str, 
        default="./output", 
        metavar="",
        help="Path to the file where the results will be saved, default: './output' "
    )
    return parser


def calculate_accuracy(all_cell_data, ground_truth):
    return np.mean(all_cell_data == ground_truth)

def calculate_rmse(all_cell_data, ground_truth):
    all_cell_data_sum = all_cell_data.apply(lambda x : x.str.split('|').str[0]).apply(pd.to_numeric, errors = "coerce") 
    + all_cell_data.apply(lambda x : x.str.split('|').str[1]).apply(pd.to_numeric, errors = "coerce")
    ground_truth_sum = ground_truth.apply(lambda x : x.str.split('|').str[0]).apply(pd.to_numeric, errors = "coerce") 
    + ground_truth.apply(lambda x : x.str.split('|').str[1]).apply(pd.to_numeric, errors = "coerce")

    return np.sqrt(mean_squared_error(ground_truth_sum ,all_cell_data_sum))

def evaluate_haplotype_predictions(haplotype, all_cell_data_path, ground_truth_path):
    """
    Evaluate the performance of a haplotype prediction tool, including RMSE, Spearman correlation coefficient, and accuracy.

    Parameters:
    haplotype (str): Specifies the haplotype mode, can be 'combined' or other values.
    all_cell_data_path (str): Path to the CSV file containing predicted data.
    ground_truth_path (str): Path to the CSV file containing ground truth data.

    Returns:
    tuple: A tuple containing the following metrics:
        - rmse (float): Root Mean Square Error (RMSE).
        - scc (float): Spearman correlation coefficient.
        - acc (float): Accuracy.
    """
    all_cell_data = pd.read_csv(all_cell_data_path, index_col=0)
    ground_truth = pd.read_csv(ground_truth_path, index_col=0)
    
    # Align rows (index) by keeping only the common indices
    common_index = all_cell_data.index.intersection(ground_truth.index)
    all_cell_data_aligned = all_cell_data.loc[common_index]
    ground_truth_aligned = ground_truth.loc[common_index]

    # Align columns by keeping only the common column names
    common_columns = all_cell_data_aligned.columns.intersection(ground_truth_aligned.columns)
    all_cell_data_aligned = all_cell_data_aligned[common_columns]
    ground_truth_aligned = ground_truth_aligned[common_columns]

    if haplotype == 'combined':
        rmse = calculate_rmse(ground_truth_aligned, all_cell_data_aligned)
    else:
        rmse = np.sqrt(mean_squared_error(ground_truth_aligned, all_cell_data_aligned))

    scc, _ = spearmanr(ground_truth_aligned.values.flatten(), all_cell_data_aligned.values.flatten())

    acc = calculate_accuracy(all_cell_data_aligned.values, ground_truth_aligned.values)
    
    return rmse, scc, acc


##  python src/main.py cndetect 
    # --tool-cna-files data/chisel_2x_500kb_combined_cnv.csv data/chisel_5x_500kb_combined_cnv.csv 
    # --cna-profile-file data/profile_combined_cnv_500kb.csv 
    # --tool-names chisel_2x_500kb chisel_5x_500kb

def run(args):
    # Validate input
    if len(args.tool_cna_files) != len(args.tool_names):
        raise ValueError("The number of classification files must match the number of tool names.")
    
    os.makedirs(args.output_dir, exist_ok=True)

    results = []
    for tool_cna_path, tool in zip(args.tool_cna_files, args.tool_names):
        all_cell_data_path = tool_cna_path
        ground_truth_path = args.cna_profile_file
        haplotype = "combined"

        rmse, scc, acc = evaluate_haplotype_predictions(haplotype, all_cell_data_path, ground_truth_path)
        
        results.append({
            "Tool":  tool,
            "RMSE": rmse,
            "SCC": scc,
            "ACC": acc
        })
    
    result_df = pd.DataFrame(results)
    output_file = os.path.join(args.output_dir, "bin_level_results.csv")
    result_df.to_csv(output_file, index=False)
    print(f"Results saved to {output_file}")



# commands/hconsetacc.py
import numpy as np
import pandas as pd
import os


def add_hconsetacc_subparser(subparsers):
    """
    Add the hconsetacc subcommand to the CLI tool.

    This subcommand is used to identify whether CNA events are correctly detected at their first branch of evolution, using ACC.
    """
    parser = subparsers.add_parser(
        "hconsetacc",
        help="Identify whether CNA events are correctly detected at their first branch of evolution, using ACC."
    )
    parser.add_argument(
        "--tool-maternal-cna-files", 
        type=str, 
        nargs = "+",
        required=True, 
        metavar="",
        help="Paths to the maternal CNA files generated by different tools. , separated by spaces"
    )
    parser.add_argument(
        "--tool-paternal-cna-files", 
        type=str, 
        nargs = "+",
        required=True, 
        metavar="",
        help="Paths to the paternal CNA files generated by different tools. , separated by spaces"
    )
    parser.add_argument(
        "--tool-names", 
        type=str, 
        nargs = "+",
        required=True, 
        metavar="",
        help="List of tool names corresponding to the classification files"
    )
    parser.add_argument(
        "--changes-file", 
        type=str, 
        required=True, 
        metavar="",
        help="Path to the changes file that contains information about changes in CN status."
    )
    parser.add_argument(
        "--output-dir", 
        type=str, 
        default="./output", 
        metavar="",
        help="Path to the file where the results will be saved, default: './output' "
    )
    return parser


def calculate_predictions_by_haplotype(change_data, predict_data, haplotype_type):

    haplotype_truth = change_data[change_data['Haplotype'] == haplotype_type]
    
    predict_sorted = predict_data.merge(haplotype_truth, left_on='region', right_on='Segment', how='inner')

    if predict_sorted.empty:
        print("Predict sorted is empty.")
        return pd.DataFrame(columns= ['Parent', 'Child', 'Haplotype', 'Type', 'Segment', 'Change', 'Parent_predict_num', 'Child_predict_num'])

    def calculate_mode_for_columns(row, columns_prefix, df):
        columns = [col for col in df.columns if col.startswith(row[columns_prefix])]
        if columns:
            values = df.loc[row.name, columns]
            mode_value = values.mode()
            return mode_value.iloc[0] if not mode_value.empty else None
        else:
            return None

    predict_sorted['Parent_predict_num'] = predict_sorted.apply(
        calculate_mode_for_columns, axis=1, columns_prefix='Parent', df=predict_sorted
    )
    
    predict_sorted['Child_predict_num'] = predict_sorted.apply(
        calculate_mode_for_columns, axis=1, columns_prefix='Child', df=predict_sorted
    )

    columns_to_keep = ['Parent', 'Child', 'Haplotype', 'Type', 'Segment', 'Change', 'Parent_predict_num', 'Child_predict_num']
    predict_sorted_cleaned = predict_sorted[columns_to_keep].copy() 

    predict_sorted_cleaned.loc[:, 'Change_predict'] = (
        predict_sorted_cleaned['Parent_predict_num'].fillna(0).astype(int).astype(str) +
        "->" +
        predict_sorted_cleaned['Child_predict_num'].fillna(0).astype(int).astype(str)
    )

    return predict_sorted_cleaned


def calculate_Evolution_Onset_accuracy(ground_truth,predict_data):
    return np.mean(predict_data == ground_truth)


##python src/main.py hconsetacc 
# --tool-maternal-cna-files data/chisel_5x_maternal_cnv.csv 
# --tool-paternal-cna-files data/chisel_5x_paternal_cnv.csv 
# --tool-names chisel_5x --changes-file profile/changes.csv

def run(args):
    # Validate input
    if (len(args.tool_maternal_cna_files) != len(args.tool_names) ) and (len(args.tool_maternal_cna_files) != len(args.tool_paternal_cna_files)):
        raise ValueError("The number of classification files must match the number of tool names.")
    
    os.makedirs(args.output_dir, exist_ok=True)
    
    change_ground_truth = pd.read_csv(args.changes_file)

    results = []
    for tool_maternal_path,tool_paternal_path, tool_name in zip(args.tool_maternal_cna_files, args.tool_paternal_cna_files,args.tool_names):

  
        predict_maternal = pd.read_csv(tool_maternal_path)
        predict_paternal = pd.read_csv(tool_paternal_path)

        maternal_result = calculate_predictions_by_haplotype(change_ground_truth, predict_maternal, haplotype_type='maternal')
        paternal_result = calculate_predictions_by_haplotype(change_ground_truth, predict_paternal, haplotype_type='paternal')

        combined_result = pd.concat([maternal_result, paternal_result], ignore_index=True)

        for type in combined_result['Type'].unique().tolist():


            acc = calculate_Evolution_Onset_accuracy(combined_result[combined_result['Type']==type]['Change'], combined_result[combined_result['Type']==type]['Change_predict'])

            results.append({
                "Tool": tool_name,
                "Type": type,
                "ACC": acc
            })
    
    result_df = pd.DataFrame(results)
    output_file = os.path.join(args.output_dir, "evolution_onset_acc.csv")
    result_df.to_csv(output_file, index=False)
    print(f"Results saved to {output_file}")



import os
import sys
import random
import datetime
import subprocess as sp
import pandas as pd
import numbers
from collections import defaultdict

# check part
def check_exist(**params):
    """Check that files are exist as expected

    Raises
    ------
    ValueError : unacceptable choice of parameters
    """
    for p in params:
        if not os.path.exists(params[p]):
            raise ValueError(
                "{} file or directory {} does not exist.".format(p, params[p]))

def check_positive(**params):
    """Check that parameters are positive as expected

    Raises
    ------
    ValueError : unacceptable choice of parameters
    """
    for p in params:
        if params[p] <= 0:
            raise ValueError(
                "Expected {} > 0, got {}".format(p, params[p]))

def check_lt_zero(**params):
    """Check that parameters are larger than zero as expected

    Raises
    ------
    ValueError : unacceptable choice of parameters
    """
    for p in params:
        if params[p] < 0:
            raise ValueError(
                "Expected {} >= 0, got {}".format(p, params[p]))


def check_int(**params):
    """Check that parameters are integers as expected

    Raises
    ------
    ValueError : unacceptable choice of parameters
    """
    for p in params:
        if not isinstance(params[p], numbers.Integral):
            raise ValueError(
                "Expected {} integer, got {}".format(p, params[p]))


def check_bool(**params):
    """Check that parameters are bools as expected

    Raises
    ------
    ValueError : unacceptable choice of parameters
    """
    for p in params:
        if params[p] is not True and params[p] is not False:
            raise ValueError(
                "Expected {} boolean, got {}".format(p, params[p]))


def check_between(v_min, v_max, **params):
    """Checks parameters are in a specified range

    Parameters
    ----------

    v_min : float, minimum allowed value (inclusive)

    v_max : float, maximum allowed value (inclusive)

    params : object
        Named arguments, parameters to be checked

    Raises
    ------
    ValueError : unacceptable choice of parameters
    """
    for p in params:
        if params[p] < v_min or params[p] > v_max:
            raise ValueError("Expected {} between {} and {}, "
                             "got {}".format(p, v_min, v_max, params[p]))

def check_in(choices, **params):
    """Checks parameters are in a list of allowed parameters
    Parameters
    ----------
    choices : array-like, accepted values
    params : object
        Named arguments, parameters to be checked
    Raises
    ------
    ValueError : unacceptable choice of parameters
    """
    for p in params:
        if params[p] not in choices:
            raise ValueError(
                "{} value {} not recognized. Choose from {}".format(
                    p, params[p], choices))

def randomSNPList(chrom_sizes, snp_ratio):
    snpList = {}
    for chrom, chrom_len in chrom_sizes.items():
        snpList[chrom] = {snp : random.sample(['A','T','C','G'], 2) for snp in random.sample(range(1, chrom_len+1), int(round(chrom_len * snp_ratio)))}
    return snpList

def parseSNPList(snpfile):
    snpList = {}
    with open(snpfile, 'r') as input:
        for line in input:
            info = line.strip().split('\t')
            chrom = info[0]
            pos = info[1]
            allele1 = info[2].upper()
            allele2 = info[3].upper()
            if allele1 not in ['A', 'C', 'G', 'T'] or allele2 not in ['A', 'C', 'G', 'T']:
                continue
            if chrom not in snpList:
                snpList[chrom] = {}
            snpList[chrom][pos] = (allele1, allele2)
    return snpList

def parseIgnoreList(ignorefile):
    ignorelist = []
    with open(ignorefile, 'r') as input:
        for line in input:
            if line != '':
                ignorelist.append(line.strip())
    return ignorelist

def random_cnv():
    # 定义数字和对应的概率e
    numbers = [0, 2, 3, 4, 5]
    probabilities = [0.2, 0.3, 0.25, 0.15, 0.1]

    # 使用choices函数进行随机选择，weights参数指定概率
    return random.choices(numbers, weights=probabilities)[0]

def random_mirrored_cnv():
    # 定义数字和对应的概率e
    numbers = [2, 3, 4, 5, 6, 7, 8, 9]
    probabilities = [0.05, 0.2, 0.3, 0.2, 0.1, 0.05, 0.05, 0.05]

    # 使用choices函数进行随机选择，weights参数指定概率
    return random.choices(numbers, weights=probabilities)[0]

def random_WGD():
    # 定义数字和对应的概率e
    numbers = [2, 3, 4, 5]
    probabilities = [0.5, 0.3, 0.15, 0.05]

    # 使用choices函数进行随机选择，weights参数指定概率
    return random.choices(numbers, weights=probabilities)[0]

def random_CNL():
    # 定义数字和对应的概率e
    numbers = [1, 2, 3, 4, 5]
    probabilities = [0.4, 0.3, 0.15, 0.1, 0.05]

    # 使用choices函数进行随机选择，weights参数指定概率
    return random.choices(numbers, weights=probabilities)[0]

def assign_cells_to_clones(cell_no, clone_no):
    # # Ensure at least one cell for each clone
    # clones = [1] * clone_no
    # remaining_cells = cell_no - clone_no

    # # Assign remaining cells to clones
    # for _ in range(remaining_cells):
    #     clone_index = random.randint(0, clone_no - 1)
    #     clones[clone_index] += 1

    # return clones
    # Initialize a list to hold the number of cells assigned to each clone
    clones = [0] * clone_no
    
    # Distribute the cells to the clones
    for i in range(cell_no):
        # Determine which clone to assign the current cell to
        clone_index = i % clone_no
        # Increment the cell count for the determined clone
        clones[clone_index] += 1
    
    return clones

def random_sampling(m, n, k):
    # m for total reads no, n for cell no, k for per reads no in cell
    if n * k < m:
        # Randomly sample without replacement
        sampled_lines = random.sample(range(m), n * k)
    else:
        # Randomly sample with replacement
        sampled_lines = [random.randint(0, m - 1) for _ in range(n * k)]

    # Reshape the sampled lines into a 2D array
    sampled_lines_2d = [sorted(sampled_lines[i:i+k]) for i in range(0, len(sampled_lines), k)]
    
    # add three lines to each line index
    new_sampled_lines_2d = []
    for lines in sampled_lines_2d:
        temp = []
        for line in lines:
            temp += [line*4+1, line*4+2, line*4+3, line*4+4]
        new_sampled_lines_2d.append(temp)
    return new_sampled_lines_2d

def root_path():
    return os.path.dirname(os.path.abspath(__file__))

class bcolors:
    HEADER = '\033[95m'  # 用于高亮显示标题或头部信息的颜色代码，通常为浅紫色。
    OKBLUE = '\033[94m'  # 一种蓝色，可能用于表示正常或信息性的消息。
    BBLUE = '\033[96m'  # 另一种蓝色，可能用于表示警告或其他需要引起注意的信息。
    OKGREEN = '\033[92m'  # 用于表示成功或正面的消息，通常为绿色。
    WARNING = '\033[93m'  # 用于表示警告或需要注意的信息，通常为黄色。
    FAIL = '\033[91m'  # 用于表示错误或失败的信息，通常为红色。
    ENDC = '\033[0m'  # 用于重置终端文本颜色，使其恢复到默认颜色。
    BOLD = '\033[1m'  # 用于加粗文本。
    UNDERLINE = '\033[4m'  # 用于下划线


# ProgressBar 在命令行界面显示进度条，向用户提供任务完成的百分比和状态更新。
class ProgressBar:

    def __init__(self, total, length, lock=None, counter=0, verbose=False, decimals=1, fill=chr(9608),
                 prefix='Progress:', suffix='Complete'):
        """
        :param total: 任务的总步骤数。
        :param length: 进度条的长度，以字符为单位。
        :param lock: 用于线程安全的锁，如果进度条在多线程环境中使用，需要这个参数。
        :param counter: 当前任务完成的步骤数，默认为0。
        :param verbose: 是否显示详细的信息，默认为False。
        :param decimals: 进度百分比的小数位数。
        :param fill: 进度条中填充部分的字符，默认为 █ ,这是一个标准的块状字符
        :param prefix: 进度条前缀，默认为"Progress:"。
        :param suffix: 进度条后缀，默认为"Complete"。
        """
        self.total = total
        self.length = length
        self.decimals = decimals
        self.fill = fill
        self.prefix = prefix
        self.suffix = suffix
        self.lock = lock
        self.counter = counter
        assert lock is not None or counter == 0
        self.verbose = verbose

    def progress(self, advance=True, msg=""):
        """
        更新进度条的状态。
        :param advance: 是否增加计数器，进度增加，默认为True。
        :param msg: 要显示的附加信息。
        """
        if self.lock is None:
            self.progress_unlocked(advance, msg)
        else:
            self.progress_locked(advance, msg)
        return True

    def progress_unlocked(self, advance, msg):
        """
        更新进度条的状态，不使用锁。
        :param advance: 是否增加计数器，进度增加，默认为True。
        :param msg: 要显示的附加信息。
        :return:
        """
        flush = sys.stderr.flush
        write = sys.stderr.write
        if advance:
            self.counter += 1
        percent = ("{0:." + str(self.decimals) + "f}").format(100 * (self.counter / float(self.total)))
        filledLength = int(self.length * self.counter // self.total)
        bar = self.fill * filledLength + '-' * (self.length - filledLength)
        rewind = '\x1b[2K\r'
        result = f'{self.prefix} |{bar}| {percent}% {self.suffix}'
        msg = f'[{datetime.datetime.now():%Y-%b-%d %H:%M:%S}] {msg}'
        if not self.verbose:
            toprint = rewind + result + f" [{msg}]"
        else:
            toprint = rewind + msg + "\n" + result
        write(toprint)
        flush()
        if self.counter == self.total:
            write("\n")
            flush()

    def progress_locked(self, advance, msg):
        """
        更新进度条的状态，*使用锁。
        :param advance: 是否增加计数器，进度增加，默认为True。
        :param msg: 要显示的附加信息。
        :return:
        """
        flush = sys.stderr.flush
        write = sys.stderr.write
        # 1. 更新操作
        if advance:
            # 1.1.使用 self.counter.get_lock() 获取锁，确保线程安全地更新计数器。
            with self.counter.get_lock():
                # 1.2.进度条的当前进度
                self.counter.value += 1
        # 2.计算进度条的完成百分比，使用格式化字符串将结果保留到指定的小数位数。
        percent = ("{0:." + str(self.decimals) + "f}").format(100 * (self.counter.value / float(self.total)))
        # 3.计算进度条中填充部分的长度，使用整数除法 // 将计数器的值除以总步骤数，然后乘以进度条的长度。
        filledLength = int(self.length * self.counter.value // self.total)
        # 4.bar 根据 filledLength 构建进度条的字符串表示：
        # 使用 self.fill 指定的字符填充已完成部分，使用 '-' 表示未完成部分。
        bar = self.fill * filledLength + '-' * (self.length - filledLength)
        # # 5.rewind 是一个 ANSI 转义序列，用于清除当前行并返回行首，这样每次更新进度条时都能在同一行显示
        rewind = '\x1b[2K\r'
        # 6.result 构建进度条的字符串，包括前缀、进度条本身、完成百分比和后缀。
        # msg 格式化消息字符串，包括当前的时间戳和传入的 msg 参数。
        result = f'{self.prefix} |{bar}| {percent}% {self.suffix}'
        msg = f'[{datetime.datetime.now():%Y-%b-%d %H:%M:%S}] {msg}'
        # 7.根据 self.verbose 的值，决定是否在进度条前显示额外的详细信息。
        # 如果False，只显示进度条和进度条提示消息；如果True，先显示工作函数的处理消息，然后是进度条。
        if not self.verbose:
            toprint = rewind + result + f" [{msg}]"
        else:
            toprint = rewind + msg + "\n" + result
        # 8.使用 self.lock 确保在写入和刷新时不会与其他线程冲突。
        with self.lock:
            sys.stderr.write(toprint)
            sys.stderr.flush()
            # 9.如果 self.counter.value 等于 self.total，表示任务已经完成。
            # 在任务完成后，输出一个换行符并刷新标准错误流，以确保进度条后面没有残留的内容。
            if self.counter.value == self.total:
                write("\n")
                flush()

# def error(msg):
#     log(msg=msg, level="ERROR")
#     sys.exit(0)

def which(program):
    """
    查找并返回一个可执行文件的路径。如果找到了，它就返回该文件的路径；如果没有找到，它就返回None。
    :param program:
    :return:
    """
    import os
    def is_exe(fpath):
        # 检查给定的文件路径是否存在并且具有执行权限。
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)

    # 提供了完整路径（fpath不为空）
    if fpath:
        # 该路径下的文件是否存在且可执行
        if is_exe(program):
            return program
    else:
        # 函数会搜索环境变量PATH中列出的所有目录
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            # 寻找第一个匹配且可执行的文件。
            if is_exe(exe_file):
                return exe_file

    return None

def runcmd(cmd, log):
    '''
    simple version of run command
    run shell command and write error log
    '''
    with open(log, 'a') as errlog:
        try:
            # Run the command
            proc = sp.Popen(
                cmd,
                shell=True,  # Allows the use of shell-specific syntax
                stdout=sp.PIPE,  # Captures standard output
                stderr=errlog  # Redirects standard error to the log file
            )
            proc.wait()  # Wait for the process to complete
            
        except Exception as e:
            print(f"Error executing command: {cmd}\n{e}")

def read_phase(allele_phase_file):
    # Use defaultdict to create a dictionary where each key maps to another dictionary
    phased = {}
    
    # Read the allele phase file use pandas
    phased_df = pd.read_csv(allele_phase_file, header=None)
    for index, row in phased_df.iterrows():
        chrom, pos, phase = row[0], row[1], row[4]
        phased[(chrom, pos)] = phase
    
    return phased

def process_snp_count_file(snp_count_file, phased_snps):
    # process the SNP count file
    snps_counts = {}
    with open(snp_count_file, 'r') as input:
        for line in input:
            fields = line.split("\t")
            chrom, pos, ref, alt, ad = fields[:5]
            pos = int(pos)
            ad = list(map(int, ad.split(",")))  # Convert AD values to integers

            # Skip positions not in phased SNPs
            if (chrom, pos) not in phased_snps:
                continue

            # Determine A and B alleles based on the phase
            phase = phased_snps[(chrom, pos)]
            a_count, b_count = 0, 0

            if phase == "0|1":
                # A allele is REF, B allele is ALT
                a_count = ad[0]
                b_count = ad[1] if len(ad) > 1 else 0
            elif phase == "1|0":
                # A allele is ALT, B allele is REF
                a_count = ad[1] if len(ad) > 1 else 0
                b_count = ad[0]

            # Store counts
            snps_counts[(chrom, pos)] = (a_count, b_count)
    return snps_counts

def write_snp_bed_file(snp_bed_file, phased_snps):
    # Write the phased SNPs to a BED file
    with open(snp_bed_file, 'w') as output:
        for snp, phase in phased_snps.items():
            chrom, pos = snp
            # phased snp is 1-based and the BED file must use 0-based start positions and 1-based end positions
            output.write('{}\t{}\t{}\n'.format(chrom, pos-1, pos))

def assign_bin(row, ref):
    # Find the bin where the SNP belongs
    bin_row = ref[
        (ref['Chromosome'] == row['chrom']) & 
        (ref['Start'] <= row['pos']) & 
        (ref['End'] >= row['pos'])
    ]
    if not bin_row.empty:
        return f"{row['chrom']}:{bin_row.iloc[0]['Start']}-{bin_row.iloc[0]['End']}"
    return None

def merge_count_files(count_files, phased_snps, ref):
    # merge count files to a matrix, row index is chrom-pos, column index is cell
    a_allele_counts = []
    b_allele_counts = []
    for count_file in count_files:
        # get cell name from path  of count file
        cell = os.path.basename(count_file).split(".")[0]
        snps_counts = process_snp_count_file(count_file, phased_snps)
        for snp, counts in snps_counts.items():
            chrom, pos = snp
            index = f"{chrom}-{pos}"
            a_allele_counts.append((chrom, pos, cell, counts[0]))
            b_allele_counts.append((chrom, pos, cell, counts[1]))
    a_allele_counts = pd.DataFrame(a_allele_counts, columns=['chrom', 'pos', 'cell', 'count'])
    b_allele_counts = pd.DataFrame(b_allele_counts, columns=['chrom', 'pos', 'cell', 'count'])

    a_allele_counts['bin'] = a_allele_counts.apply(assign_bin, axis=1, ref=ref)

    # Step 4: Group by bins and cells, summing the counts
    a_allele_counts_grouped = a_allele_counts.groupby(['bin', 'cell'])['count'].sum().reset_index()
    b_allele_counts_grouped = b_allele_counts.groupby(['bin', 'cell'])['count'].sum().reset_index()

    # Step 5: Pivot to create the count matrix
    a_allele_martrix = a_allele_counts_grouped.pivot(index='bin', columns='cell', values='count')
    b_allele_martrix = b_allele_counts_grouped.pivot(index='bin', columns='cell', values='count')

    # Step 6: Fill missing values with 0
    a_allele_martrix = a_allele_martrix.fillna(0)
    b_allele_martrix = b_allele_martrix.fillna(0)

    baf_matrix = b_allele_martrix / (a_allele_martrix + b_allele_martrix)
    return a_allele_martrix, b_allele_martrix, baf_matrix

def extract_chr_and_start(bin_value):
    chrom, positions = bin_value.split(":")
    start, _ = positions.split("-")
    
    # extract chromosome
    if chrom.startswith("chr"):
        chrom = chrom[3:]
    chrom = int(chrom) if chrom.isdigit() else chrom
    
    # extract start position
    start = int(start)
    
    return chrom, start

def merge_coverage_files(coverage_files):
    merged_df = pd.DataFrame()
    # merge coverage files to a matrix, row index is chrom-pos, column index is cell
    for cov_file in coverage_files:
        cell_name = os.path.basename(cov_file).split(".")[0]
        
        df = pd.read_csv(cov_file, sep="\t", header=None, names=["chr", "start", "end", "coverage"])
        df["start"] =  df["start"] + 1
        df["bin"] = df["chr"] + ":" + df["start"].astype(str) + "-" + df["end"].astype(str)
        
        df = df[["bin", "coverage"]].set_index("bin")
        df.rename(columns={"coverage": cell_name}, inplace=True)
        
        if merged_df.empty:
            merged_df = df
        else:
            merged_df = merged_df.join(df, how="outer")
    
    merged_df = merged_df.sort_values(by="bin", key=lambda col: col.map(extract_chr_and_start))
    return merged_df

def create_allele_count_matrices(count_files, phased_snps_filtered, a_allele_bed, b_allele_bed):
    """
    Process SNP count files and create TSV matrices for A and B allele counts.
    
    Args:
        count_files: List of paths to count files
        phased_snps_filtered: Dictionary or set of filtered phased SNPs
        output_dir: Directory to save output files
    
    Returns:
        tuple: Paths to the created A and B allele count matrices
    """
    # Initialize dictionaries to store counts for each allele
    a_allele_counts = defaultdict(dict)
    b_allele_counts = defaultdict(dict)
    
    # Process each count file
    with open(a_allele_bed, 'w') as a_allele, open(b_allele_bed, 'w') as b_allele:
        for count_file in count_files:
            # Extract cell name from filename
            cell = os.path.basename(count_file).split(".")[0]
            
            # Process the count file to get SNP counts
            snps_counts = process_snp_count_file(count_file, phased_snps_filtered)
            
            # Store counts in the appropriate dictionaries
            for snp, counts in snps_counts.items():
                chrom, pos = snp
                snp_id = f"{chrom}:{pos}"
                
                # Store A allele count
                a_allele_counts[snp_id][cell] = counts[0]
                
                # Store B allele count
                b_allele_counts[snp_id][cell] = counts[1]
                a_allele.write('\t'.join([chrom, str(pos), str(pos+1), cell, str(counts[0])]) + '\n')
                b_allele.write('\t'.join([chrom, str(pos), str(pos+1), cell, str(counts[1])]) + '\n')
    
    # Convert dictionaries to pandas DataFrames
    a_allele_df = pd.DataFrame.from_dict(a_allele_counts, orient='index')
    b_allele_df = pd.DataFrame.from_dict(b_allele_counts, orient='index')
    
    # Fill NaN values with 0 (for cells with no counts at a particular SNP)
    a_allele_df = a_allele_df.fillna(0).astype(int)
    b_allele_df = b_allele_df.fillna(0).astype(int)
    
    # Sort the rows by chromosome and position
    def sort_key(snp_id):
        chrom, pos = snp_id.split(':')
        # Handle chromosome sorting (numeric, then X, Y, etc.)
        if chrom.startswith('chr'):
            chrom = chrom[3:]
        
        if chrom.isdigit():
            chrom_val = int(chrom)
        elif chrom == 'X':
            chrom_val = 100
        elif chrom == 'Y':
            chrom_val = 101
        else:
            chrom_val = 999
            
        return (chrom_val, int(pos))
    
    # Sort the DataFrames
    a_allele_df = a_allele_df.loc[sorted(a_allele_df.index, key=sort_key)]
    b_allele_df = b_allele_df.loc[sorted(b_allele_df.index, key=sort_key)]
    
    return a_allele_df, b_allele_df


import argparse
from hcdsim import HCDSIM
import inspect

def create_hcdsim_from_args(args):
    """
    from args buid HCDSIM
    """
    hcdsim_params = inspect.signature(HCDSIM.__init__).parameters

    # filter None value and same params
    hcdsim_args = {
        key: value
        for key, value in vars(args).items()
        if key in hcdsim_params and value is not None
    }

    # create HCDSim class
    return HCDSIM(**hcdsim_args)

def sim(args):
    # run hcdsim full pipeline
    hcdsim = create_hcdsim_from_args(args)
    hcdsim.sim()

def generate_profile(args):
    hcdsim = create_hcdsim_from_args(args)
    hcdsim.gprofile()

def generate_fasta(args):
    hcdsim = create_hcdsim_from_args(args)
    hcdsim.gfasta()

def generate_fastq(args):
    hcdsim = create_hcdsim_from_args(args)
    hcdsim.gfastq()

def alignment(args):
    hcdsim = create_hcdsim_from_args(args)
    hcdsim.align()

def downsampling(args):
    hcdsim = create_hcdsim_from_args(args)
    hcdsim.downsam()

def process_cell_bam(args):
    hcdsim = create_hcdsim_from_args(args)
    hcdsim.pbam()

def generate_barcode_bam(args):
    hcdsim = create_hcdsim_from_args(args)
    hcdsim.bcbam()

def cal_rdr(args):
    hcdsim = create_hcdsim_from_args(args)
    hcdsim.rdr()

def cal_baf(args):
    hcdsim = create_hcdsim_from_args(args)
    hcdsim.baf()

def add_shared_arguments(parser):
    # Basics, Inputs and Executables
    parser.add_argument('-r', '--ref-genome', type=str, metavar="", help='Path to reference genome [required]')
    parser.add_argument('-o', '--outdir', type=str, required=False, default='./', metavar="", help='Output directory (default: current directory)')
    parser.add_argument('-g', '--ignore', type=str, required=False, default=None, metavar="", help='Path to the exclusion list of contigs file containing a line-per-line list with chromosome names to ignore in the reference (default: none)')
    parser.add_argument('-b', '--bin-size', type=str, required=False, default='5Mb', metavar="", help='The fixed bin size, with or without \"kb\" or \"Mb\" (default: 5Mb)')
    parser.add_argument('-cno', '--clone-no', type=int, required=False, default=3, metavar="", help='The random clone number contained in evolution tree, including normal clone (default: 3)')
    parser.add_argument('-eno', '--cell-no', type=int, required=False, default=5, metavar="", help='The total cell number for this simultion dataset (deafult: 5)')
    parser.add_argument('-t', '--thread', type=int, required=False, default=None, metavar="", help='Number of parallele jobs to use (default: equal to number of available processors)')
    
def add_executable_arguments(parser):
    parser.add_argument('--wgsim', type=str, required=False, default='wgsim', metavar="", help='Path to the executable \"wgsim\" fileexecutable (default: in $PATH)')
    parser.add_argument('--bwa', type=str, required=False, default='bwa', metavar="", help='Path to the executable \"bwa\" file (default: in $PATH)')
    parser.add_argument('--samtools', type=str, required=False, default='samtools', metavar="", help='Path to the executable \"samtools\" file (default: in $PATH)')
    # parser.add_argument('--picard', type=str, required=False, default='picard.jar', metavar="", help='Path to the \"picard.jar\" file (default: ./picard.jar)')

def add_gprofile_arguments(parser):
    # CNA Profile
    parser.add_argument('-d', '--max-tree-depth', type=int, required=False, default=4, metavar="", help='The maximum depth of random evolution tree (default: 4)')
    parser.add_argument('-cp', '--cna-prob-cutoff', type=float, required=False, default=0.8, metavar="", help='The cutoff probability of a bin undergoing CNA, if random probability is larger than cutoff, CNA happens (default: 0.8)')
    parser.add_argument('-wgd', '--wgd-cna-no', type=int, required=False, default=0, metavar="", help='Number of clonal whole-genome duplications (WGDs) to introduce in the ancestor of tumor evolution (default: 0)')
    parser.add_argument('-wcl', '--wcl-cna-no', type=int, required=False, default=0, metavar="", help='Number of clonal whole-chromosome losses (WCLs) to introduce in the ancestor of tumor evolution (default: 0)')
    parser.add_argument('-loh', '--loh-cna-no', type=int, required=False, default=30, metavar="", help='Number of loss of heterozygosity (LOH) CNAs to introduce in the each clone of tumor evolution, including CNL_LOH, CNN_LOH, CNG_LOH (default: 30)')
    parser.add_argument('-goh', '--goh-cna-no', type=int, required=False, default=10, metavar="", help='Number of gain of heterozygosity (GOH) CNAs to introduce in the each clone of tumor evolution (default: 10)')
    parser.add_argument('-m', '--mirror-cna-no', type=int, required=False, default=10, metavar="", help='Number of mirror CNAs to introduce in the each clone of tumor evolution (default: 10)')
    parser.add_argument('-l', '--snp-list', type=str, required=False, default=None, metavar="", help='Path to the known germline SNPs file containing a SNP positions to add in the simulate human genome in the\nformat "#CHR POSITION REF_ALLELES ALT_ALLELES" with first line as headline (default: none, SNPs are placed randomly)')
    parser.add_argument('-p', '--snp-ratio', type=float, required=False, default=0.001, metavar="", help='Ratio of SNPs to place randomly when a snp file is not given (snp-ratio is requried only whether snp-list is not provided. default: 0.001).')
    parser.add_argument('-hr', '--heho-ratio', type=float, required=False, default=0.67, metavar="", help='Ratio of heterozygous SNPs compared to homozygous ones (default: 0.67)')

def add_gfasta_arguments(parser):
    # Clone FASTA
    pass
    
def add_gfastq_arguments(parser):
    # Clone FASTQ
    parser.add_argument('-c', '--clone-coverage', type=float, required=False, default=30, metavar="", help='The reads coverage for clone (default: 30)')
    parser.add_argument('-rl', '--reads-len', type=int, required=False, default=150, metavar="", help='The length of the reads in FASTQ (default: 150)')
    parser.add_argument('-i', '--insertion-size', type=int, required=False, default=350, metavar="", help='The outer distance between the two ends (default: 350)')
    parser.add_argument('-e', '--error-rate', type=float, required=False, default=0.0, metavar="", help='The base error rate (default: 0.0)')

def add_align_arguments(parser):
    # Alignment
    pass
    

def add_downsample_arguments(parser):
    # Downsampling
    parser.add_argument('-cc', '--cell-coverage', type=float, required=False, default=0.5, metavar="", help='The reads coverage for clone (default: 0.5)')

def add_pbam_arguments(parser):
    # Process Cell Bam
    pass

def main():
    parser = argparse.ArgumentParser(
        prog="hcdsim",
        description="HCDSIM: A Single-Cell Genomics Simulator with Haplotype-Specific Copy Number Annotation",
    )

    # Add subcommands
    subparsers = parser.add_subparsers(
        title="Subcommands",
        description="Available subcommands",
        help="Use `hcdsim <subcommand> --help` for more information.",
        dest="subcommand"
    )
    subparsers.required = True

    # sim subcommand
    sim_parser = subparsers.add_parser("sim", help="Running the complete HCDSIM pipeline.")
    add_shared_arguments(sim_parser)
    add_gprofile_arguments(sim_parser)
    add_gfasta_arguments(sim_parser)
    add_gfastq_arguments(sim_parser)
    add_align_arguments(sim_parser)
    add_downsample_arguments(sim_parser)
    add_pbam_arguments(sim_parser)
    add_executable_arguments(sim_parser)
    sim_parser.set_defaults(func=sim)

    # gprofile subcommand
    gprofile_parser = subparsers.add_parser("gprofile", help="Generating CNA profile.")
    add_shared_arguments(gprofile_parser)
    add_gprofile_arguments(gprofile_parser)
    gprofile_parser.set_defaults(func=generate_profile)

    # gfasta subcommand
    gfasta_parser = subparsers.add_parser("gfasta", help="Generating clone FASTA file.")
    add_shared_arguments(gfasta_parser)
    add_gfasta_arguments(gfasta_parser)
    gfasta_parser.set_defaults(func=generate_fasta)

    # gfastq subcommand
    gfastq_parser = subparsers.add_parser("gfastq", help="Generating clone FASTQ file.")
    add_shared_arguments(gfastq_parser)
    add_gfastq_arguments(gfastq_parser)
    gfastq_parser.add_argument('--wgsim', type=str, required=False, default='wgsim', metavar="", help='Path to the executable \"wgsim\" fileexecutable (default: in $PATH)')
    gfastq_parser.set_defaults(func=generate_fastq)

    # align subcommand
    align_parser = subparsers.add_parser("align", help="Aligning clone FASTQ file.")
    add_shared_arguments(align_parser)
    add_align_arguments(align_parser)
    align_parser.add_argument('--bwa', type=str, required=False, default='bwa', metavar="", help='Path to the executable \"bwa\" file (default: in $PATH)')
    align_parser.add_argument('--samtools', type=str, required=False, default='samtools', metavar="", help='Path to the executable \"samtools\" file (default: in $PATH)')
    align_parser.set_defaults(func=alignment)

    # downsample subcommand
    downsample_parser = subparsers.add_parser("downsam", help="Downsampling clone BAM.")
    add_shared_arguments(downsample_parser)
    add_downsample_arguments(downsample_parser)
    downsample_parser.add_argument('--samtools', type=str, required=False, default='samtools', metavar="", help='Path to the executable \"samtools\" file (default: in $PATH)')
    downsample_parser.set_defaults(func=downsampling)

    # pbam subcommand
    pbam_parser = subparsers.add_parser("pbam", help="Processing cell BAMs.")
    add_shared_arguments(pbam_parser)
    add_pbam_arguments(pbam_parser)
    pbam_parser.add_argument('--samtools', type=str, required=False, default='samtools', metavar="", help='Path to the executable \"samtools\" file (default: in $PATH)')
    # pbam_parser.add_argument('--picard', type=str, required=False, default='picard.jar', metavar="", help='Path to the \"picard.jar\" file (default: in ./)')
    pbam_parser.set_defaults(func=process_cell_bam)

    # bcbam subcommand
    bcbam_parser = subparsers.add_parser("bcbam", help="Generating barcode BAM file.")
    add_shared_arguments(bcbam_parser)
    bcbam_parser.add_argument('-bcl', '--barcode-len', type=int, required=False, default=12, metavar="", help='Length of barcodes (default: 12)')
    bcbam_parser.add_argument('--bwa', type=str, required=False, default='bwa', metavar="", help='Path to the executable \"bwa\" file (default: in $PATH)')
    bcbam_parser.add_argument('--samtools', type=str, required=False, default='samtools', metavar="", help='Path to the executable \"samtools\" file (default: in $PATH)')
    bcbam_parser.add_argument('--bcftools', type=str, required=False, default='bcftools', metavar="", help='Path to the executable \"bcftools\" file (default: in $PATH)')
    bcbam_parser.set_defaults(func=generate_barcode_bam)

    # rdr subcommand
    rdr_parser = subparsers.add_parser("rdr", help="Computing RDRs.")
    add_shared_arguments(rdr_parser)
    rdr_parser.add_argument('--bedtools', type=str, required=False, default='bedtools', metavar="", help='Path to the executable \"bedtools\" file (default: in $PATH)')
    rdr_parser.set_defaults(func=cal_rdr)

    # baf subcommand
    baf_parser = subparsers.add_parser("baf", help="Computing BAFs.")
    add_shared_arguments(baf_parser)
    baf_parser.add_argument('--bcftools', type=str, required=False, default='bcftools', metavar="", help='Path to the executable \"bcftools\" file (default: in $PATH)')
    baf_parser.add_argument('--bedtools', type=str, required=False, default='bedtools', metavar="", help='Path to the executable \"bedtools\" file (default: in $PATH)')
    baf_parser.set_defaults(func=cal_baf)


    # 解析命令行参数
    args = parser.parse_args()
    if args.ref_genome is None or args.ref_genome.strip() == '': #or not os.path.isfile(args.ref_genome):
        raise ValueError("The specified human reference-genome file is required and does not exist!")
    args.func(args)

if __name__ == "__main__":
    main()


# commands/hccnchange.py
import numpy as np
import pandas as pd
import os

from sklearn.metrics import mean_squared_error


def add_hccnchange_subparser(subparsers):
    """
    Add the hccnchange subcommand to the CLI tool.

    This subcommand is used to identify if the CN change from parent to child is correctly identified, using RMSE and ACC.
    """
    parser = subparsers.add_parser(
        "hccnchange",
        help="Identify if the CN change from parent to child is correctly identified, using RMSE and ACC."
    )
    parser.add_argument(
        "--tool-maternal-cna-files", 
        type=str, 
        nargs = "+",
        required=True, 
        metavar="",
        help="Paths to the maternal CNA files generated by different tools. , separated by spaces"
    )
    parser.add_argument(
        "--tool-paternal-cna-files", 
        type=str, 
        nargs = "+",
        required=True, 
        metavar="",
        help="Paths to the paternal CNA files generated by different tools. , separated by spaces"
    )
    parser.add_argument(
        "--tool-names", 
        type=str, 
        nargs = "+",
        required=True, 
        metavar="",
        help="List of tool names corresponding to the classification files"
    )
    parser.add_argument(
        "--changes-file", 
        type=str, 
        required=True, 
        metavar="",
        help="Path to the changes file that contains information about changes in CN status."
    )
    parser.add_argument(
        "--output-dir", 
        type=str, 
        default="./output", 
        metavar="",
        help="Path to the file where the results will be saved, default: './output' "
    )
    return parser


def calculate_predictions_by_haplotype(change_data, predict_data, haplotype_type):

    haplotype_truth = change_data[change_data['Haplotype'] == haplotype_type]
    
    predict_sorted = predict_data.merge(haplotype_truth, left_on='region', right_on='Segment', how='inner')

    if predict_sorted.empty:
        print("Predict sorted is empty.")
        return pd.DataFrame(columns= ['Parent', 'Child', 'Haplotype', 'Type', 'Segment', 'Change', 'Parent_predict_num', 'Child_predict_num'])

    def calculate_mode_for_columns(row, columns_prefix, df):
        columns = [col for col in df.columns if col.startswith(row[columns_prefix])]
        if columns:
            values = df.loc[row.name, columns]
            mode_value = values.mode()
            return mode_value.iloc[0] if not mode_value.empty else None
        else:
            return None

    predict_sorted['Parent_predict_num'] = predict_sorted.apply(
        calculate_mode_for_columns, axis=1, columns_prefix='Parent', df=predict_sorted
    )
    
    predict_sorted['Child_predict_num'] = predict_sorted.apply(
        calculate_mode_for_columns, axis=1, columns_prefix='Child', df=predict_sorted
    )

    columns_to_keep = ['Parent', 'Child', 'Haplotype', 'Type', 'Segment', 'Change', 'Parent_predict_num', 'Child_predict_num']
    predict_sorted_cleaned = predict_sorted[columns_to_keep].copy() 

    predict_sorted_cleaned.loc[:, 'Change_predict'] = (
        predict_sorted_cleaned['Parent_predict_num'].fillna(0).astype(int).astype(str) +
        "->" +
        predict_sorted_cleaned['Child_predict_num'].fillna(0).astype(int).astype(str)
    )

    return predict_sorted_cleaned

def calculate_CN_Change_accuracy(ground_truth,predict_data):
    parent = ground_truth.str.split('->').str[0].apply(pd.to_numeric, errors='coerce')
    parent_predict = predict_data.str.split('->').str[0].apply(pd.to_numeric, errors='coerce')

    child = ground_truth.str.split('->').str[1].apply(pd.to_numeric, errors='coerce')
    child_predict = predict_data.str.split('->').str[1].apply(pd.to_numeric, errors='coerce')
    return np.mean((parent-child) == (parent_predict-child_predict))

def calculate_CN_change_rmse(ground_truth,predict_data):
    parent = ground_truth.str.split('->').str[0].apply(pd.to_numeric, errors='coerce')
    parent_predict = predict_data.str.split('->').str[0].apply(pd.to_numeric, errors='coerce')

    child = ground_truth.str.split('->').str[1].apply(pd.to_numeric, errors='coerce')
    child_predict = predict_data.str.split('->').str[1].apply(pd.to_numeric, errors='coerce')

    rmse = np.sqrt(mean_squared_error(parent-child, parent_predict-child_predict))
    
    return rmse

# python src/main.py hccnchange /
# --tool-maternal-cna-files data/chisel_5x_maternal_cnv.csv /
# --tool-paternal-cna-files data/chisel_5x_paternal_cnv.csv /
# --tool-names chisel_5x --changes-file profile/changes.csv /

def run(args):
    # Validate input
    if (len(args.tool_maternal_cna_files) != len(args.tool_names) ) and (len(args.tool_maternal_cna_files) != len(args.tool_paternal_cna_files)):
        raise ValueError("The number of classification files must match the number of tool names.")
    
    os.makedirs(args.output_dir, exist_ok=True)
    
    change_ground_truth = pd.read_csv(args.changes_file)

    results = []
    for tool_maternal_path,tool_paternal_path, tool_name in zip(args.tool_maternal_cna_files, args.tool_paternal_cna_files,args.tool_names):

  
        predict_maternal = pd.read_csv(tool_maternal_path)
        predict_paternal = pd.read_csv(tool_paternal_path)

        maternal_result = calculate_predictions_by_haplotype(change_ground_truth, predict_maternal, haplotype_type='maternal')
        paternal_result = calculate_predictions_by_haplotype(change_ground_truth, predict_paternal, haplotype_type='paternal')

        combined_result = pd.concat([maternal_result, paternal_result], ignore_index=True)

        for type in combined_result['Type'].unique().tolist():


            rmse = calculate_CN_change_rmse(combined_result[combined_result['Type']==type]['Change'], combined_result[combined_result['Type']==type]['Change_predict'])
            acc = calculate_CN_Change_accuracy(combined_result[combined_result['Type']==type]['Change'], combined_result[combined_result['Type']==type]['Change_predict'])

            results.append({
                "Tool": tool_name,
                "Type": type,
                "RMSE": rmse,
                "ACC": acc
            })
    
    result_df = pd.DataFrame(results)
    output_file = os.path.join(args.output_dir, "evolution_onset_CN_Change.csv")
    result_df.to_csv(output_file, index=False)
    print(f"Results saved to {output_file}")



# commands/cnclass.py

import pandas as pd
import numpy as np
import os
from sklearn.metrics import auc, precision_recall_curve, roc_auc_score, average_precision_score, accuracy_score, precision_score, recall_score, f1_score, confusion_matrix, cohen_kappa_score, brier_score_loss

import warnings


warnings.filterwarnings("ignore")

def add_cnclass_subparser(subparsers):
    """
    Add the cnclass subcommand to the CLI tool.

    This subcommand is used to calculate CN state detection metrics at the bin level.
    """
    parser = subparsers.add_parser(
        "cnclass",
        help="Calculate CN state detection metrics at the bin level."
    )
    parser.add_argument(
        "--tool-maternal-cna-files", 
        type=str, 
        nargs = "+",
        required=True,
        metavar="FILE",
        help="Paths to the maternal/minor CNA files generated by different tools. , separated by spaces"
    )
    parser.add_argument(
        "--tool-paternal-cna-files", 
        type=str, 
        nargs = "+",
        required=True,
        metavar="FILE",
        help="Paths to the paternal/major CNA files generated by different tools. , separated by spaces"
    )
    parser.add_argument(
        "--tool-names", 
        type=str, 
        nargs = "+",
        required=True,
        metavar="NAME",
        help="List of tool names corresponding to the CNA files."
    )
    parser.add_argument(
        "--profile-maternal-cna-files", 
        type=str, 
        required=True,
        metavar="FILE",
        help="Path to the maternal/minor ground truth CNA file."
    )
    parser.add_argument(
        "--profile-paternal-cna-files", 
        type=str, 
        required=True,
        metavar="FILE",
        help="Path to the paternal/major ground truth CNA file."
    )
    parser.add_argument(
        "--type", 
        type=str, 
        choices=["acCNA", "hcCNA"],
        required=True,
        metavar="TYPE",
        help="Type of CNA to analyzs. Choose between 'acCNA' and 'hcCNA'."
    )
    parser.add_argument(
        "--output-dir", 
        type=str, 
        default="./output", 
        metavar="DIR",
        help="Path to the file where the results will be saved, default: './output' "
    )
    return parser

def categorize_and_save_cnv(ground_truth_file, predict_file, save_dir,tool, condition,type = "maternal"):
    """
    Categorize and save CNVs based on a specific condition.

    Parameters:
        ground_truth_file (str): Path to the ground truth file.
        predict_file (str): Path to the prediction file.
        save_dir (str): Directory to save categorized results.     
        tool (str): Tool name.
        condition (str): Condition for categorization (e.g., '>=2').
        cna_type (str): Type of CNA .
    """
    ground_truth_df = pd.read_csv(ground_truth_file)
    predict_df = pd.read_csv(predict_file)
    
    def categorize_cn(value):
        if condition == ">=2":
            return int(value >= 2)
        
        if condition == ">=9":
            return int(value >= 9)
        

        if condition.startswith("="):
            try:
                threshold = float(condition[1:])  
                return int(value == threshold)
            except ValueError:
                raise ValueError(f"Invalid threshold in condition: {condition}")
        raise ValueError(f"Unsupported condition: {condition}")
    
    
    ground_truth_categorized = ground_truth_df.copy()
    predict_categorized = predict_df.copy()

    for col in ground_truth_df.columns[1:]:  
        ground_truth_categorized[col] = ground_truth_df[col].apply(categorize_cn)
        predict_categorized[col] = predict_df[col].apply(categorize_cn)

    os.makedirs(save_dir, exist_ok=True)
    ground_truth_categorized.to_csv(os.path.join(save_dir, f'ground_truth_{type}.csv'), index=False)
    predict_categorized.to_csv(os.path.join(save_dir, f'{tool}_predict_{type}.csv'), index=False)


def align_dataframes(ground_truth_df, predict_df):
    """
    Align two dataframes by their common columns and index.

    Parameters:
        ground_truth_df (DataFrame): Ground truth dataframe.
        predict_df (DataFrame): Predicted values dataframe.

    Returns:
        Tuple[DataFrame, DataFrame]: Aligned ground truth and predicted dataframes.
    """
    common_columns = ground_truth_df.columns.intersection(predict_df.columns)
    ground_truth_df = ground_truth_df[common_columns]
    predict_df = predict_df[common_columns]
    
    common_index = ground_truth_df.index.intersection(predict_df.index)
    ground_truth_df = ground_truth_df.loc[common_index]
    predict_df = predict_df.loc[common_index]
    
    return ground_truth_df, predict_df


def calculate_metrics(ground_truth_df, predict_df):
    """
    Calculate evaluation metrics for CNV detection.

    Parameters:
        ground_truth_df (DataFrame): Ground truth values.
        predict_df (DataFrame): Predicted values.

    Returns:
        dict: A dictionary containing evaluation metrics.
    """
    y_true = ground_truth_df.values.flatten()  
    y_pred = predict_df.values.flatten()  

    unique_y_true = set(y_true)
    unique_y_pred = set(y_pred)
    
    metrics = {}
    
    if len(set(y_true)) > 1 and len(set(y_pred)) > 1:
        metrics['AUROC'] = roc_auc_score(y_true, y_pred)
        precision, recall, _ = precision_recall_curve(y_true, y_pred)
        metrics['AUPRC'] = auc(recall, precision)
    else:
        metrics['AUROC'] = None
        metrics['AUPRC'] = None

    metrics['Accuracy'] = accuracy_score(y_true, y_pred)
    metrics['Precision'] = precision_score(y_true, y_pred, zero_division=0)
    metrics['Recall'] = recall_score(y_true, y_pred, zero_division=0)
    metrics['F1 Score'] = f1_score(y_true, y_pred, zero_division=0)
    metrics['Kappa'] = cohen_kappa_score(y_true, y_pred)
    metrics['Brier Score'] = brier_score_loss(y_true, y_pred)

    unique_labels = sorted(unique_y_true.union(unique_y_pred))

    # Check if y_true and y_pred contain more than one class before calculating confusion matrix
    if len(set(y_true)) > 1 and len(set(y_pred)) > 1:
        # Calculate confusion matrix
        tn, fp, fn, tp = confusion_matrix(y_true, y_pred,labels=unique_labels).ravel()
        metrics['Specificity'] = tn / (tn + fp) if (tn + fp) != 0 else None
        metrics['PPV'] = metrics['Precision']  # PPV is the same as Precision in binary classification
        metrics['NPV'] = tn / (tn + fn) if (tn + fn) != 0 else None
    else:
        # If we cannot calculate confusion matrix due to lack of classes
        metrics['Specificity'] = None
        metrics['PPV'] = None
        metrics['NPV'] = None

    return metrics


def process_files_in_folder(folder_path,tool,haplotype_list):
    results = {}
    for file_type in haplotype_list:
        ground_truth_file = os.path.join(folder_path, f'ground_truth_{file_type}.csv')
        predict_file = os.path.join(folder_path, f'{tool}_predict_{file_type}.csv')
        
        if os.path.exists(ground_truth_file) and os.path.exists(predict_file):
            ground_truth_df = pd.read_csv(ground_truth_file)
            predict_df = pd.read_csv(predict_file)

            ground_truth_df, predict_df = align_dataframes(ground_truth_df, predict_df)

            results[file_type] = {}
            
            for clone in predict_df.columns[1:].str.split('_').str[0].unique():
                ground_truth_subclone_df = ground_truth_df.loc[:, ground_truth_df.columns.str.split('_').str[0] == clone]
                predict_subclone_df = predict_df.loc[:, predict_df.columns.str.split('_').str[0] == clone]

                metrics = calculate_metrics(ground_truth_subclone_df, predict_subclone_df)

                results[file_type][clone] = metrics

    return results

def save_metrics_to_csv(results, folder_name, folder_path):
    for file_type,clone_metrics in results.items():
        all_metrics = []
        for clone, metrics in clone_metrics.items():
            metrics_with_info = {
                'Clone': clone
            }
            metrics_with_info.update(metrics)
            all_metrics.append(metrics_with_info)

        metrics_df = pd.DataFrame(all_metrics)
        save_path = os.path.join(folder_path, f'metrics_{folder_name}_{file_type}.csv')
        metrics_df.to_csv(save_path, index=False)



def run(args):
     # Validate input
    if (len(args.tool_maternal_cna_files) != len(args.tool_names) ) and (len(args.tool_maternal_cna_files) != len(args.tool_paternal_cna_files)):
        raise ValueError("The number of classification files must match the number of tool names.")
    
    if args.type not in ["acCNA", "hcCNA"]:
        raise ValueError("The type must be either 'acCNA' or 'hcCNA'.")
    if args.type == "acCNA":
        haplotype_list = ['minor', 'major']
    elif args.type == "hcCNA":
        haplotype_list = ['maternal', 'paternal']
    
    
    profile_maternal_path = args.profile_maternal_cna_files
    profile_paternal_path = args.profile_paternal_cna_files
    
    conditions_and_dirs = [
        ('>=2', 'CN_Gain'),
        ('=1', 'CN_Neutral'),
        ('=0', 'CN_Loss'),
        ('=2', 'CN_equal_2'),
        ('=3', 'CN_equal_3'),
        ('=4', 'CN_equal_4'),
        ('=5', 'CN_equal_5'),
        ('=6', 'CN_equal_6'),
        ('=7', 'CN_equal_7'),
        ('=8', 'CN_equal_8'),
        ('>=9', 'CN_over_9'),
    ]
    
    os.makedirs(args.output_dir, exist_ok=True)

    all_metrics = []
    for tool_maternal_path,tool_paternal_path, tool_name in zip(args.tool_maternal_cna_files, args.tool_paternal_cna_files,args.tool_names):
        for condition, save_dir in conditions_and_dirs:
            categorize_and_save_cnv(
                ground_truth_file=profile_maternal_path,  
                predict_file=tool_maternal_path,  
                save_dir=f"{args.output_dir}/{tool_name}/{save_dir}",
                tool=tool_name,
                condition=condition,
                type=haplotype_list[0]
            )

        for condition, save_dir in conditions_and_dirs:
            categorize_and_save_cnv(
                ground_truth_file= profile_paternal_path,  
                predict_file=tool_paternal_path,  
                save_dir=f"{args.output_dir}/{tool_name}/{save_dir}",
                tool=tool_name,
                condition=condition,
                type=haplotype_list[1]
            )  

        for condition, save_dir in conditions_and_dirs:
            results = process_files_in_folder(f"{args.output_dir}/{tool_name}/{save_dir}",tool_name,haplotype_list)
            for file_type, clone_metrics in results.items():
                for clone, metrics in clone_metrics.items():
                    metrics_with_info = {
                        'Type': save_dir,
                        'Haplotype': file_type,
                        'Clone': clone,
                        'tool':tool_name
                    }
                    metrics_with_info.update(metrics)  
                    
                    all_metrics.append(metrics_with_info)
    
    result_df = pd.DataFrame(all_metrics)
    output_file = os.path.join(args.output_dir, "cnclass_results.csv")
    result_df.to_csv(output_file, index=False)
    print(f"Results saved to {output_file}")



# commands/mirrorsubclone.py

import numpy as np
import pandas as pd
from sklearn.metrics import adjusted_rand_score as adjustedRandIndex
from sklearn.metrics import adjusted_mutual_info_score as AMI
import os


def add_mirrorsubclone_subparser(subparsers):
    """
    Add the mirrorsubclone subcommand to the CLI tool.

    This subcommand is used to calculate the RMSE,ACC of Mirror-subclone CNA.
    """
    parser = subparsers.add_parser(
        "mirrorsubclone",
        help="Calculate the RMSE,ACC of Mirror-subclone CNA."
    )
    parser.add_argument(
        "--tool-cna-files", 
        type=str, 
        nargs = "+",
        required=True, 
        metavar="",
        help="Paths to the CNA files generated by different tools."
    )
    parser.add_argument(
        "--tool-names", 
        type=str, 
        nargs = "+",
        required=True, 
        metavar="",
        help="List of tool names corresponding to the classification files"
    )
    parser.add_argument(
        "--changes-file", 
        type=str, 
        required=True, 
        metavar="",
        help="Path to the changes file that contains information about changes in CN status."
    )
    parser.add_argument(
        "--output-dir", 
        type=str, 
        default="./output", 
        metavar="",
        help="Path to the file where the results will be saved, default: './output' "
    )
    return parser

def process_mirror_change(change_file_path):

    change_ground_truth = pd.read_csv(change_file_path)

    change_ground_truth['Segment'] = (
        change_ground_truth['Chromosome'] + ":" +
        change_ground_truth['Start'].astype(str) + "-" +
        change_ground_truth['End'].astype(str)
    )

    return change_ground_truth


def calculate_predictions(change_data, predict_data):
    predict_sorted = predict_data.merge(change_data, left_on='region', right_on='Segment', how='inner')

    if predict_sorted.empty:
        print("Predict sorted is empty.")
        return pd.DataFrame(columns= ['Segment','Clone1', 'Clone2', 'Clone1_CNA','Clone2_CNA', 'Clone1_predict_CNA', 'Clone2_predict_CNA'])

    def calculate_mode_for_columns(row, columns_prefix, df):
        columns = [col for col in df.columns if col.startswith(row[columns_prefix])]
        if columns:
            values = df.loc[row.name, columns]
            mode_value = values.mode()
            return mode_value.iloc[0] if not mode_value.empty else None
        else:
            return None

    predict_sorted['Clone1_predict_CNA'] = predict_sorted.apply(
        calculate_mode_for_columns, axis=1, columns_prefix='Clone1', df=predict_sorted
    )
    
    predict_sorted['Clone2_predict_CNA'] = predict_sorted.apply(
        calculate_mode_for_columns, axis=1, columns_prefix='Clone2', df=predict_sorted
    )

    columns_to_keep = ['Segment','Clone1', 'Clone2', 'Clone1_CNA','Clone2_CNA', 'Clone1_predict_CNA', 'Clone2_predict_CNA']
    predict_sorted_cleaned = predict_sorted[columns_to_keep]
    
    return predict_sorted_cleaned


##python src/main.py mirrorsubclone --tool-cna-files data/chisel_5x_combined_cnv.csv --tool-names chisel_5x --changes-file profile/mirrored_subclonal_cnas.csv

def run(args):
    # Validate input
    if len(args.tool_cna_files) != len(args.tool_names):
        raise ValueError("The number of classification files must match the number of tool names.")
    
    os.makedirs(args.output_dir, exist_ok=True)

    change_data = process_mirror_change(args.changes_file)

    results = []
    for tool_path, tool_name in zip(args.tool_cna_files, args.tool_names):

        predict_combined = pd.read_csv(tool_path)

        combined_result = calculate_predictions(change_data, predict_combined)

        combined_result['Clone1_maternal_CNA'] = combined_result['Clone1_CNA'].str.split('|').str[0]
        combined_result['Clone1_paternal_CNA'] = combined_result['Clone1_CNA'].str.split('|').str[1]
        combined_result['Clone2_maternal_CNA'] = combined_result['Clone2_CNA'].str.split('|').str[0]
        combined_result['Clone2_paternal_CNA'] = combined_result['Clone2_CNA'].str.split('|').str[1]

        combined_result['Clone1_predict_maternal_CNA'] = combined_result['Clone1_predict_CNA'].str.split('|').str[0]
        combined_result['Clone1_predict_paternal_CNA'] = combined_result['Clone1_predict_CNA'].str.split('|').str[1]
        combined_result['Clone2_predict_maternal_CNA'] = combined_result['Clone2_predict_CNA'].str.split('|').str[0]
        combined_result['Clone2_predict_paternal_CNA'] = combined_result['Clone2_predict_CNA'].str.split('|').str[1]

        combined_result['acc_result'] = (
            (combined_result['Clone1_predict_maternal_CNA'].astype(str) == combined_result['Clone1_maternal_CNA'].astype(str) )  &
            (combined_result['Clone1_predict_paternal_CNA'].astype(str) ==combined_result['Clone1_paternal_CNA'].astype(str) ) &
            (combined_result['Clone2_predict_maternal_CNA'].astype(str) == combined_result['Clone2_maternal_CNA'].astype(str) )  &
            (combined_result['Clone2_predict_paternal_CNA'].astype(str) ==combined_result['Clone2_paternal_CNA'].astype(str) ) 
        )

        combined_result['Clone1_maternal_error'] = (
            (combined_result['Clone1_predict_maternal_CNA'].astype(float) - combined_result['Clone1_maternal_CNA'].astype(float))**2
        )

        combined_result['Clone1_paternal_error'] = (
            (combined_result['Clone1_predict_paternal_CNA'].astype(float) - combined_result['Clone1_paternal_CNA'].astype(float))**2
        )

        combined_result['Clone2_maternal_error'] = (
            (combined_result['Clone2_predict_maternal_CNA'].astype(float) - combined_result['Clone2_maternal_CNA'].astype(float))**2
        )

        combined_result['Clone2_paternal_error'] = (
            (combined_result['Clone2_predict_paternal_CNA'].astype(float) - combined_result['Clone2_paternal_CNA'].astype(float))**2
        )

        combined_result['RMSE_result'] = (
            np.sqrt(combined_result['Clone1_maternal_error']) +
            np.sqrt(combined_result['Clone1_paternal_error']) +
            np.sqrt(combined_result['Clone2_maternal_error']) +
            np.sqrt(combined_result['Clone2_paternal_error'])
        )

        rmse = combined_result['RMSE_result'].mean()

        acc = combined_result['acc_result'].mean()

        results.append({
            "Tool": tool_name,
            "RMSE": rmse,
            "ACC": acc
        })
    
    result_df = pd.DataFrame(results)
    output_file = os.path.join(args.output_dir, "mirror_subclone_result.csv")
    result_df.to_csv(output_file, index=False)
    print(f"Results saved to {output_file}")



# commands/hcPhasing.py

import pandas as pd
import os

def add_hcPhasing_subparser(subparsers):
    """
    Add the hcPhasing subcommand to the CLI tool.

    This subcommand is used to calculate the mismatch error and switch error.
    
    """
    parser = subparsers.add_parser(
        "hcPhasing",
        help="Calculate Mismatch Error and Switch Error."
    )
    parser.add_argument(
        "--tool-maternal-cna-files", 
        type=str, 
        nargs = "+",
        required=True, 
        help="Paths to the maternal CNA files generated by different tools. , separated by spaces"
    )
    parser.add_argument(
        "--tool-paternal-cna-files", 
        type=str, 
        nargs = "+",
        required=True, 
        help="Paths to the paternal CNA files generated by different tools. , separated by spaces"
    )
    parser.add_argument(
        "--ground-truth-maternal-file", 
        type=str, 
        required=True, 
        help="Path to the ground truth maternal file."
    )
    parser.add_argument(
        "--ground-truth-paternal-file", 
        type=str, 
        required=True, 
        help="Path to the ground truth paternal file."
    )
    parser.add_argument(
        "--tool-names", 
        type=str, 
        nargs = "+",
        required=True, 
        help="List of tool names corresponding to the classification files"
    )
    parser.add_argument(
        "--output-dir", 
        type=str, 
        default="./output", 
        help="Path to the file where the results will be saved, default: './output' "
    )
    return parser

def generate_binary_df(tool_maternal_df,tool_paternal_df):
    """
    Generate binary DataFrames for maternal and paternal data.
    """
    maternal_binary_df = (tool_maternal_df > tool_paternal_df).astype(int)
    maternal_binary_df = maternal_binary_df.mask(tool_maternal_df == tool_paternal_df, -1)
    maternal_binary_df = maternal_binary_df.mask(tool_maternal_df.isna() | tool_paternal_df.isna())

    paternal_binary_df = (tool_paternal_df > tool_maternal_df).astype(int)
    paternal_binary_df = paternal_binary_df.mask(tool_maternal_df == tool_paternal_df, -1)
    paternal_binary_df = paternal_binary_df.mask(tool_maternal_df.isna() | tool_paternal_df.isna())

    return maternal_binary_df, paternal_binary_df


def evaluate_mismatch_and_switch_error(ground_maternal_binary_df,ground_paternal_binary_df,maternal_binary_df,paternal_binary_df, tool_name="tool_name"):

    ground_maternal_binary_df = ground_maternal_binary_df.dropna(axis=1, how='all')
    ground_paternal_binary_df = ground_paternal_binary_df.dropna(axis=1, how='all')

    common_index = ground_maternal_binary_df.index.intersection(maternal_binary_df.index)
    common_columns = ground_maternal_binary_df.columns.intersection(maternal_binary_df.columns)

    ground_maternal_binary_df = ground_maternal_binary_df.loc[common_index, common_columns]
    ground_paternal_binary_df = ground_paternal_binary_df.loc[common_index, common_columns]

    maternal_binary_df = maternal_binary_df.loc[common_index, common_columns]
    paternal_binary_df = paternal_binary_df.loc[common_index, common_columns]

    column_prefixes = maternal_binary_df.columns.str.extract(r"(^[^_]+)_")[0].unique()

    result_list = []

    for prefix in column_prefixes:

        group_columns = [col for col in maternal_binary_df.columns if col.startswith(f"{prefix}_")]

        gm = ground_maternal_binary_df[group_columns]
        gp = ground_paternal_binary_df[group_columns]
        m = maternal_binary_df[group_columns]
        p = paternal_binary_df[group_columns]

        maternal_mask = gm != -1
        paternal_mask = gp != -1

        maternal_diff_count = (m != gm)[maternal_mask].sum().sum()
        paternal_diff_count = (p != gp)[paternal_mask].sum().sum()

        total_maternal = maternal_mask.sum().sum()
        total_paternal = paternal_mask.sum().sum()


        if maternal_diff_count != paternal_diff_count:
            print(f"{tool_name} mismatch maternal and paternal result not same!")

        result = {
            "tool_name": tool_name,
            "cell_group": prefix,
            "mismatch_count": maternal_diff_count,
            "total": total_maternal,
            "mismatch_ratio": maternal_diff_count / total_maternal if total_maternal > 0 else None,
        }

        #-----------------switch error-----------------

        n_row, n_col =  gm.shape

        total = 0
        maternal_diff_count = 0
        paternal_diff_count = 0
        for row in range(0,n_row):
            for col in range(0,n_col-1):
                if gm.iloc[row,col] == -1 or gm.iloc[row,col+1] == -1:
                    continue
                else:
                    total +=1
                    if gm.iloc[row,col] != m.iloc[row,col] or gm.iloc[row,col+1] != m.iloc[row,col+1]:
                        maternal_diff_count +=1

                    if gp.iloc[row,col] != p.iloc[row,col] or gp.iloc[row,col+1] != p.iloc[row,col+1]:
                        paternal_diff_count +=1


        if maternal_diff_count != paternal_diff_count:
            print(f"{tool_name} switch error maternal and paternal result not same!")

        result.update({
            "switch_error_count": maternal_diff_count,
            "total_switch_compare_count": total,
            "switch_error_ratio": maternal_diff_count / total if total > 0 else None,
        })

        result_list.append(result)
            
    
    return result_list

# python src/main.py hcPhasing --tool-maternal-cna-files data/chisel_5x_maternal_cnv.csv 
# --tool-paternal-cna-files data/chisel_5x_paternal_cnv.csv 
# --ground-truth-maternal-file data/ground_truth_maternal_cnv.csv 
# --ground-truth-paternal-file data/ground_truth_paternal_cnv.csv --tool-names chisel_5x

def run(args):
    # Validate input
    if len(args.tool_maternal_cna_files) != len(args.tool_names):
        raise ValueError("The number of classification files must match the number of tool names.")
    
    os.makedirs(args.output_dir, exist_ok=True)

    ground_maternal_df = pd.read_csv(args.ground_truth_maternal_file, index_col=0)
    ground_paternal_df = pd.read_csv(args.ground_truth_paternal_file, index_col=0)

    ground_maternal_binary_df, ground_paternal_binary_df = generate_binary_df(ground_maternal_df, ground_paternal_df)

    results = []
    for maternal_path,paternal_path, tool in zip(args.tool_maternal_cna_files,args.tool_paternal_cna_files, args.tool_names):
        tool_maternal_df = pd.read_csv(maternal_path, index_col=0)
        tool_paternal_df = pd.read_csv(paternal_path, index_col=0)

        maternal_binary_df, paternal_binary_df = generate_binary_df(tool_maternal_df, tool_paternal_df)

        result_list = evaluate_mismatch_and_switch_error(ground_maternal_binary_df, ground_paternal_binary_df,maternal_binary_df, paternal_binary_df, tool)
        
        results.extend(result_list)
    
    result_df = pd.DataFrame(results)
    output_file = os.path.join(args.output_dir, "hcPhasing.csv")
    result_df.to_csv(output_file, index=False)
    print(f"Results saved to {output_file}")
